#!/usr/bin/perl

# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# It is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# To receive a copy of the GNU General Public License, please write to the
# Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

# Creates authority records (a marcxml file) from biblio records
# Stefano Bargioni, Pontificia Università della Santa Croce, Rome - 2012-02-29

# With Koha ILS, act as follows:
# select marc from biblioitems into outfile 'bibs.mrc' fields escaped by '';
# Edit create_authorities_from_bibs.xml to fit your needings.
# ./create_authorities_from_bibs.pl [-c config_file.xml] bibs.mrc > authorities.mrcxml
# Default for -c is create_authorities_from_bibs.xml
# bulkmarcimport.pl -a -m marcxml -file authorities.mrcxml
# link_bibs_to_authorities.pl

use strict;
use MARC::Batch;
# use Encode qw(encode decode);
use utf8;
use open qw( :std :utf8);
use IO::File;
use Getopt::Long qw(:config no_ignore_case);
use Pod::Usage;
use XML::Simple;
use Date::Format;

my $VERSION = '1.2';

my $help;
my $xml_config_file;
my $version = '';
GetOptions(
    'h'   => \$help,
    'c:s' => \$xml_config_file,
    'V'   => \$version
) or pod2usage(2);

if ($version) {
    print "$VERSION\n";
    exit 1;
}
pod2usage( -exitstatus => 0, -verbose => 2 ) if ( $help || $#ARGV == -1 );

die("File not found ".$ARGV[$#ARGV] ) unless ( -e $ARGV[$#ARGV]);

my $config=(XML::Simple->new()->XMLin($xml_config_file)); # read config from "create_authorities.xml" or -c filename

my $authorities = $config->{'authorities'}; # array reference
my $LIBRARY_LANGUAGE_CODE = $config->{'library_language_code'};
my $LIBRARY_CODE = $config->{'library_code'};
my @lt = localtime(time); 
my $today = strftime("%y%m%d", @lt);
my $now = strftime("%Y%m%d%H%M%S.0", @lt);
my $YYMMDD = UNIVERSAL::isa($config->{'yymmdd'}, 'HASH') ? $today : $config->{'yymmdd'}; # (use XML::Simple option ForceContent?)
my $CURRENT_TIMESTAMP = UNIVERSAL::isa($config->{'current_timestamp'}, 'HASH') ? $now : $config->{'current_timestamp'};

tail_xml_models($authorities);

print '<?xml version="1.0" encoding="UTF-8"?>
<collection
  xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xsi:schemaLocation="http://www.loc.gov/MARC21/slim
http://www.loc.gov/standards/marcxml/schema/MARC21slim.xsd"
  xmlns="http://www.loc.gov/MARC21/slim">'."\n";

for (my $i = 0; $i < scalar @$authorities; $i++) {
	print STDERR $$authorities[$i]->{'code'}.": ";
	my %values; # this can cause high RAM consumption
	my $n = extract_values($$authorities[$i],\%values);
	my $m = generate_marcxml($$authorities[$i],\%values);
	print STDERR "$m/$n"."\n";
}

print "</collection>\n";

# -------------

sub tail_xml_models {
	my $authorities = shift;
	
	for (my $i = 0; $i < scalar @$authorities; $i++) {
		$$authorities[$i]->{'xml_model'} =~ s/\*LIBRARY_LANGUAGE_CODE\*/$LIBRARY_LANGUAGE_CODE/mg;
		$$authorities[$i]->{'xml_model'} =~ s/\*LIBRARY_CODE\*/$LIBRARY_CODE/mg;
		$$authorities[$i]->{'xml_model'} =~ s/\*CURRENT_TIMESTAMP\*/$CURRENT_TIMESTAMP/mg;
		my $T008 = $$authorities[$i]->{'t008'};
		$$authorities[$i]->{'xml_model'} =~ s/\*T008\*/$T008/mg;
		$$authorities[$i]->{'xml_model'} =~ s/\*YYMMDD\*/$YYMMDD/mg;
	}
}

sub extract_values {
	my $authtype = shift;
	my $values = shift;

	my $records_read = 0;
	my $batch;
	my $fh = IO::File->new( $ARGV[$#ARGV] ); # don't let MARC::Batch open the file, as it applies the ':utf8' IO layer
	$batch = MARC::Batch->new( 'USMARC', $fh );
	$batch->warnings_off();
	$batch->strict_off(); # don't die for errors in the input file
	
	RECORD: while () {
	    my $record;
	    eval { $record = $batch->next() };
	    if ($@) {
	        print STDERR "Bad MARC record " . $records_read . ": skipped.\n";
	        next;
	    }
	    last unless ($record);

	    $records_read++;
		manage_record($record,$authtype,$values);
	}
	return $records_read;
}

sub manage_record {
	my $record = shift;
	my $authtype = shift;
	my $values = shift;

	# $authtype->{tag} can be string or array (use XML::Simple option ForceArray?)
	my $tag_list = ref($authtype->{'tag'}) eq 'ARRAY' ? $authtype->{'tag'} : [$authtype->{'tag'}];
	for (my $i = 0; $i < scalar @$tag_list; $i++) {
		my $tag = substr(@$tag_list[$i],0,3);
		# my $subfields = substr(@$tag_list[$i],3);
		# $subfields = '*' if ($subfields eq ''); # every subfield of the biblio rec will be in the auth rec, but subfield 9
		my @fields = $record->field($tag);
		for (my $j = 0; $j < scalar @fields; $j++) {
			my $field = $fields[$j];
			my $subfields = substr(@$tag_list[$i],3);
			$subfields = '*' if ($subfields eq ''); # every subfield of the biblio rec will be in the auth rec, but subfield 9
			if ($subfields eq '*') {
				# my @sfs = $record->field($tag)->subfields();
				my @sfs = $field->subfields();
				$subfields='';
				for (my $m = 0; $m < scalar @sfs; $m++) {
					$subfields .= $sfs[$m][0]; # translate * into the list of subfields for current tag
				}
				$subfields =~ s/9//g; # always eliminates subfield '9'
			}
			# eliminates multiple subfield codes (like 'axx' to 'ax')
			my $shash = {};
			for  (my $k = 0; $k < length($subfields); $k++) {
				$shash->{substr($subfields,$k,1)}=1;
			}
			$subfields=''; while( my ($k, $v) = each %$shash ) {$subfields .= $k};
			# for each required subfield
			my $value = '';
			for (my $k = 0; $k < length($subfields); $k++) {
				my $subfield = substr($subfields,$k,1);
				my @ss = $field->subfield($subfield);
				for (my $p = 0; $p < scalar @ss; $p++) {
					my $s = $ss[$p];
					if ($s && $subfield ne '9') {
						$value .= ' <subfield code="'.$subfield.'">'.entities($s).'</subfield>';
					}
				}
			}
			# $values->{$value} = '1'; # create the voice, if doesn't exist
			$values->{$value} = [$record->field($tag)->indicator(1),$record->field($tag)->indicator(2)]; # create the voice if doesn't exist, and store indicators
		}
	}
	return scalar(keys %$values);
}

sub generate_marcxml {
	my $authtype = shift;
	my $values = shift;

	foreach my $value (keys %$values) {
		my $record_xml = $authtype->{'xml_model'};
		$record_xml =~ s/\*VALUE\*/$value/mg;
		my ($ind1,$ind2) = @{$values->{$value}};
		# indicators (for MARC21: 1xx (ind1), 130 (ind2))
		$record_xml =~ s/\*IND1\*/$ind1/mg if ($authtype->{'ind1'} && $authtype->{'ind1'} eq '1');
		$record_xml =~ s/\*IND1\*/$ind2/mg if ($authtype->{'ind1'} && $authtype->{'ind1'} eq '2');
		$record_xml =~ s/\*IND2\*/$ind2/mg if ($authtype->{'ind2'} && $authtype->{'ind1'} eq '2');
		$record_xml =~ s/\*IND2\*/$ind1/mg if ($authtype->{'ind2'} && $authtype->{'ind2'} eq '1');
		$record_xml =~ s/\*IND1\*/ /mg;
		$record_xml =~ s/\*IND2\*/ /mg;
		print $record_xml;
	}
	return scalar(keys %$values);
}

sub entities {
	my $s = shift;

	$s =~ s/&/&amp;/g;
	$s =~ s/</&lt;/g;
	$s =~ s/>/&gt;/g;
	return $s;
}

=head1 create_authorities_from_bibs.pl

   Print a marcxml file to STDOUT with authority records inferred from bibliographic records.
And statistics to STDERR.

   The xml config file must be edited to correctly fit your needs.

=head1 SYNOPSIS
 
./create_authorities_from_bibs.pl [options] F<file.mrc>
 
=head1 OPTIONS
 
=over 8
 
=item B<-h>
 
Print this message and exit.
 
=item B<-c>

Specify xml config file other than ./create_authorities_from_bibs.xml

=item B<-V>

Print the version and exit.

=item B<file.mrc>

The mandatory ISO2709 file to read. Can be STDIN, '-'.

=back
 
=head1 DESCRIPTION
 
   When your library doesn't use authority records, this tool helps you to generate them starting from bibliographic records.
   A configuration file must be edited to define authority types (personal|corporative|meeting names, subjects, uniform titles,...),
bibliographic tags and subfields where to find values.
   In the configuration file you also have to set your library code, your library language code, timestamp and date required for tags 040, 005, and 008.
See the xml configuration file for more details.
   This tool can be used with any MARC flavour (MARC21, UNIMARC, ...).

   Of course, generated authority records will only include accepted forms, since 'see', 'see also', etc. are not contained in bibliographic records.

=head1 LIMITATIONS

Uniqueness of headers is done using hash. This can cause high RAM consumption depending on how many different values for each authority type are in your file.mrc.

Accepts and returns only UTF-8.

=head1 AUTHOR

Pontificia Universita' della Santa Croce <http://www.pusc.it/bib/>

Stefano Bargioni <bargioni@pusc.it>

=head1 CREDITS

Paola Manoni, Biblioteca Apostolica Vaticana

=head1 RELEASE NOTES

1.2    2012-03-30  Bug fixed for multiple instance of same field.

1.1    2012-03-29  Correct multiple instance of same subfield.

1.0    2012-03-22  Initial release.

=cut
